#include "Exception.hpp"
#include "DanuString.hpp"
#include <new>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <windows.h>

namespace Danu
{
	/**
	 * The constructor replaces the library supplied std::unexpected and std::terminate
	 * function by the Exception::unexpected() and Exception::terminate(). It also initializes
	 * the Exception::traceFunction with a default Exception::TraceFunction implementation
	 * if no other function is registered.
	 */
	Exception::ExceptionInitializer::ExceptionInitializer()
	{
		std::cerr << "std::cerr: + "<< __DANUFUNCTION__ << std::endl;
		Exception::enableDefaultTrace();
		std::set_unexpected(Danu::Exception::unexpected);
		std::set_terminate(Danu::Exception::terminate);
		std::cerr << "std::cerr: + "<< __DANUFUNCTION__ << std::endl;
	}
			
	/**
	 * The destructor deletes the Exception::traceFunction.
	 */
	Exception::ExceptionInitializer::~ExceptionInitializer()
	{
		std::cerr << "std::cerr: + "<< __DANUFUNCTION__ << std::endl;
		Exception::disableDefaultTrace();
		std::cerr << "std::cerr: + "<< __DANUFUNCTION__ << std::endl;
	}
	
	/// Linked list of exception texts
	/**
	 * The last added text will be the first in the list. If we want to print the list from the
	 * beginning we must iterate backwards.	
	*/
	class ExceptionText
	{
		public:
			ExceptionText(	const char* anExceptionText,
							const ExceptionText* aNextExceptionText);
			ExceptionText(const ExceptionText& anExceptionText);
	
			~ExceptionText()
			{
				if (exceptionText) delete [] exceptionText;
				if (nextExceptionText) delete nextExceptionText;
			}
	
			const char* getText() const
			{
				return exceptionText;
			}
	
			ExceptionText* getNext() const
			{
				return nextExceptionText;
			}
				
			void	appendText( const char* aTextToAppend) throw();
	
			char* exceptionText;
			ExceptionText* nextExceptionText;
	}; // class ExceptionText
	
	ExceptionText::ExceptionText(	const char* anExceptionText,
									const ExceptionText* aNextExceptionText)
	{
		//	We replace the new_handler supplied by the Standard Library (or someone else's) because it's
		// required behaviour is (see 18.4.22) to:
		//	1.	Make more storage available for allocation and return
		//	2.	Throw an axception pf type bad_alloc or a class derived from bad_alloc
		//	3.	Call either abort() or exit()
		//	We do not want behaviour 2 and/or 3 to occur while we are handling an exception:
		//	2.	Throwing an exception in an exception leads to a unbalanced try/catch which leads
		//		to calling terminate() (see 18.6)
		//	3.	This prevents executing the corresponding catch
		//	We loose 1 but for now we assume that it is unlikely that a shortage of memory will occur:
		//	4 Gb should be enough for any application to find some room to store these few bytes of
		//	messages ;-)
		void(*oldHandler)() = std::set_new_handler(0);
	
		//	for new(std::nothrow) see 18.4.11
		exceptionText = new(std::nothrow) char[strlen(anExceptionText)+1];
		if (exceptionText) strcpy(exceptionText, anExceptionText);
		nextExceptionText = (ExceptionText*)aNextExceptionText;
	
		//	Restore the handler
		std::set_new_handler(oldHandler);
	}
	
	ExceptionText::ExceptionText ( const ExceptionText& anExceptionText )
	{
		// See the other constructor for implementation comments
		void(*oldHandler)() = std::set_new_handler(0);
	
		exceptionText = new(std::nothrow) char[strlen(anExceptionText.exceptionText)+1];
		if (exceptionText)
		{
			strcpy(exceptionText, anExceptionText.exceptionText);
			if (anExceptionText.nextExceptionText)
			{
			  // Do a deep copy for the next ExceptionText in our list
			  // by calling it's copy constructor
			 nextExceptionText = new(std::nothrow) ExceptionText(*(anExceptionText.nextExceptionText));
			}else
			{
			 nextExceptionText = 0;
			}
		}else // new failed and now we hope for the best by just setting pointers
		{
			exceptionText = anExceptionText.exceptionText;
			nextExceptionText = anExceptionText.nextExceptionText;
		}
	
		// See the other constructor for implementation comments
		std::set_new_handler(oldHandler);
	}

	/**
	 * This function append a text entry to this text. Because this function will 
	 * be used uring exception throwing we do not allow this function to throw exceptions itself 
	 * (it would lead to calling terminate() because a catch too less). If there is not enough 
	 * memory available to add the text this function will do nothing.
	 */
	void ExceptionText::appendText(	const char* aTextToAppend) throw()
	{
		// See the constructor for implementation comments
		void(*oldHandler)() = std::set_new_handler(0);
	
		char* aText;
		if (exceptionText)
		{
			aText = new(std::nothrow) char[strlen(exceptionText)+strlen(aTextToAppend)+1];
		} else
		{
			aText = new(std::nothrow) char[strlen(aTextToAppend)+1];
		}
	
		if (aText)
		{
			if (exceptionText)
			{
				strcpy(aText, exceptionText);
				delete [] exceptionText;
				strcat(aText, aTextToAppend);
				exceptionText = aText;
			}else
			{
				strcpy(aText, aTextToAppend);
				exceptionText = aText;
			}
		}
	
		// See the constructor for implementation comments
		std::set_new_handler(oldHandler);
	}
	
//	/* static */ Exception::ExceptionInitializer Exception::exceptionInitializer;
	/* static */ Exception::TraceFunction* Exception::defaultTraceFunction = 0;
	/* static */ Exception::TraceFunction* Exception::currentTraceFunction = 0;
		
	Exception::Exception(	const char* anExceptionText,
							unsigned long anErrorId /*= 0*/,
							Severity aSeverity /*= Exception::unrecoverable */) throw()
		:	exceptionSeverity(aSeverity),
			exceptionErrorId(anErrorId),
			exceptionLocationCount(0),
			exceptionText(0)
	{
		// See the constructor of ExceptionText for implementation comments
		void(*oldHandler)() = std::set_new_handler(0);
	
		if (anExceptionText)
		{
			exceptionText = new(std::nothrow) ExceptionText(anExceptionText, 0);
			if (exceptionText)
			{
				exceptionTextLevelCount = 1;
			}
		}		
		// See the constructor of ExceptionText for implementation comments
		std::set_new_handler(oldHandler);
	}
	
	Exception::Exception(	const std::string& anExceptionText,
							unsigned long anErrorId /*= 0*/,
							Severity aSeverity /*= Exception::unrecoverable */) throw()
		:	exceptionSeverity(aSeverity),
			exceptionErrorId(anErrorId),
			exceptionLocationCount(0),
			exceptionText(0)
	{
		// See the constructor of ExceptionText for implementation comments
		void(*oldHandler)() = std::set_new_handler(0);
	
		if (anExceptionText.length() > 0)
		{
			exceptionText = new(std::nothrow) ExceptionText(anExceptionText.c_str(), 0);
			if (exceptionText)
			{
				exceptionTextLevelCount = 1;
			}
		}
		// See the constructor of ExceptionText for implementation comments
		std::set_new_handler(oldHandler);
	}
	
	Exception::Exception(const Exception& aException) throw()
		:	exceptionSeverity(aException.exceptionSeverity),
			exceptionErrorId(aException.exceptionErrorId),
			exceptionLocationCount(aException.exceptionLocationCount),
			exceptionText(0),			
			exceptionTextLevelCount(aException.exceptionTextLevelCount),
			whatText(aException.whatText)
	{
		// See the constructor of ExceptionText for implementation comments
		void(*oldHandler)() = std::set_new_handler(0);
	
		// The cpy ctor of the ExceptionText will take care of the deep copy
		if (aException.exceptionText != 0)
		{
			exceptionText = new (std::nothrow) ExceptionText(*(aException.exceptionText));
		}
		// copy the array of locations
		for (unsigned long i=0; i<aException.getLocationCount(); i++ )
		{
			exceptionLocation[i] = aException.exceptionLocation[i];
		}
	
		// See the constructor of ExceptionText for implementation comments
		std::set_new_handler(oldHandler);
	}
	
	Exception::~Exception() throw()
	{
		// The destructor of ExceptionText takes care of the linked list behaviour bij calling delete
		// for the next in the list
		if (exceptionText) delete exceptionText;
	}
	
	Exception::Severity Exception::getSeverity() const
	{
		return exceptionSeverity;
	}
	
	unsigned long Exception::getErrorId() const
	{
		return exceptionErrorId;
	}
	/**
	 * If there is no location at anIndex this function will return 0.
	 * @param anIndex must be 0-4 because we only have 5 slots for a location.
	 */
	const ExceptionLocation* Exception::getLocation(unsigned long anIndex /* = 0*/) const
	{
		if (anIndex < exceptionLocationCount)
		{
			return &(exceptionLocation[anIndex]);
		}
		return 0;
	}
	
	/**
	 * The internal location array is 0 based. To get the last location do 
	 * getLocation(getLocationCount()-1)!
	 */
	unsigned long Exception::getLocationCount() const
	{
		return exceptionLocationCount;
	}
	
	/**
	 * If there is no text at anIndex this function will return 0.
	 * The internal text linked list is 0 based. To get the last text do 
	 * getText(getTextCount()-1)!
	 */
	const char* Exception::getText(unsigned long anIndex /* = 0*/) const
	{
		ExceptionText* tmp = exceptionText;
		
		// Check for a valid index
		if (getTextCount() > anIndex)
		{
			// Advance untill the index is reached (zero counted)
			for (unsigned long i = 0; i < anIndex; ++i)
			{
				tmp = tmp->getNext();
			}
			return tmp->getText();
		} else
		{
			return 0;
		}
	}
	
	unsigned long Exception::getTextCount() const
	{
		return exceptionTextLevelCount;
	}
	
	void Exception::setSeverity(Severity aSeverity)
	{
		exceptionSeverity = aSeverity;
	}
	
	void Exception::setErrorId(unsigned long anErrorId)
	{
		exceptionErrorId = anErrorId;
	}
	/**
	 * Exception uses an array with 5 slots (0-4) as storage for locations. If a 6th location is
	 * added this funtion will do nothing.
	 */
	void Exception::addLocation(const ExceptionLocation& aExceptionLocation)
	{
		if(exceptionLocationCount < 5)
		{
			exceptionLocation[exceptionLocationCount] = aExceptionLocation;
			exceptionLocationCount++;
		}
	}
	/**
	 * This function adds a text entry to a unbounded linked list. Because this function will be used
	 * during exception throwing we do not allow this function to throw exceptions itself (it would 
	 * lead to calling terminate() because a catch too less). If there is not enough memory available
	 * to add the text this function will do nothing.
	 */
	void Exception::addText(const char* anExceptionText) throw()
	{
		if (anExceptionText!= 0)
		{
			// See the constructor of ExceptionText for implementation comments
			void(*oldHandler)() = std::set_new_handler(0);
	
			ExceptionText* currentExceptionText= exceptionText;
			exceptionText = new(std::nothrow) ExceptionText(anExceptionText, currentExceptionText);
			if (exceptionText)
			{
				exceptionTextLevelCount++;
			}else
			{
				exceptionText= currentExceptionText;
			}
	
			// Restore the original new handler
			std::set_new_handler(oldHandler);
	   }
	}
	
	/**
	 * This function append a text entry to the last added exeption text. Because this function will 
	 * be used uring exception throwing we do not allow this function to throw exceptions itself 
	 * (it would lead to calling terminate() because a catch too less). If there is not enough 
	 * memory available to add the text this function will do nothing.
	 */
	void Exception::appendText(const char* anErrorText) throw()
	{
		if (anErrorText != 0)
		{
			if (exceptionText)
			{
				exceptionText->appendText(anErrorText);
			}else
			{
				addText(anErrorText);
			}
		}
	}
		
	/* static */ void Exception::enableDefaultTrace()
	{
		if(!Exception::getTraceFunction())
		{
			Exception::setTraceFunction((Exception::defaultTraceFunction = new Exception::TraceFunction));
		}
		Exception::getTraceFunction()->trace("> Exception logging enabled\n");
	}
				
	/* static */ void Exception::disableDefaultTrace()
	{
		Exception::getTraceFunction()->trace("> Exception logging disabled\n");
		delete Exception::defaultTraceFunction;
		Exception::defaultTraceFunction = 0;
	}
	
	/* static */ void Exception::reEnableDefaultTrace()
	{
		if(!Exception::defaultTraceFunction)
		{
			Exception::setTraceFunction((Exception::defaultTraceFunction = new Exception::TraceFunction));
		}else
		{
			Exception::setTraceFunction(Exception::defaultTraceFunction);
		}
		Exception::getTraceFunction()->trace("Exception logging re-enabled by the default trace function\n");
	}
		
	/**
	 * Replaces the current function by the new one. Exception assumes NO OWNERSHIP of the 
	 * pointer and will NOT DELETE the trace function if the program terminates. 
	 * 
	 * A working trace function is installed even if a user does not set a user defined function, 
	 * or if the user calls it with 0. In that case a default Exception::TraceFunction that is
	 * created and maintained by the private exceptionInitializer object will be used. This function
	 * prints all it's output to stderr using fprintf.
	 */
	/* static */ void Exception::setTraceFunction( Exception::TraceFunction* aTraceFunction)
	{
		currentTraceFunction = aTraceFunction;
		// If the traceFunction is replaced by nothing, reinstate the default one.
		if(currentTraceFunction == 0)
		{
			// From the Exception's point of view there is always a TraceFunction
			// with the exceptionInitializer object so we re-enable.
			// This will call this function recursively for a second time, but now 
			// with a valid aTraceFunction....
			Exception::reEnableDefaultTrace();
		}
	}
	
	/**
	 * This function can return 0.
	 */
	/* static */ Exception::TraceFunction* Exception::getTraceFunction( )
	{
		return currentTraceFunction;
	}

	/**
	 The output will be something like:
	 \code
	 ExceptionName anExceptionName thrown:
	 	function: Original_function
	 	line: Original_line
	 	file: Original_file
	 	ErrorId:	
	 Exceptiontexts:
	 text 1............................
	 text 2............................
	 text n............................
	 Rethrown at:
	 	function 2, line 2, file 2
	 	function 3, line 3, file 3
	 	function 4, line 4, file 4
	 	function 5, line 5, file 5<b
	 \endcode
	 */
	const char* Exception::what() const throw()
	{
		// See the constructor of ExceptionText for implementation comments
		void(*oldHandler)() = std::set_new_handler(0);
		
		// Reset whatText  to empty
		/// \todo Is the lifetime of c_str() bug still in std::string???
		whatText = "";
	
		// Determine the largest buffer we will need for printing.
		size_t maxLength = 200;
		size_t length = 0;

		// How much for first throw (if there is a location recorded)
		static const char* throwHeaderText = "%s exception thrown:\n\tfunction: %s\n\tline: %d\n\tfile: %s\n";	
		if (exceptionLocationCount != 0)		
		{				
			length =	strlen(throwHeaderText) +
						strlen(name()) +
						strlen(exceptionLocation[0].getFunctionName()) +
						strlen(exceptionLocation[0].getFileName());
			if(length> maxLength) maxLength = length;
		}
		
		// Room for the errorId
		// max unsigned == 4294967295, ergo 10 chars
		//                 unknown is shorter
		length = strlen("\tErrorId: 0123456789\n");
		if(length> maxLength) maxLength = length;

		// Count all the exception texts (if any)
		if (exceptionTextLevelCount > 0)
		{
			// We go from the first (max getTextCount) to the last
			for (int i = exceptionTextLevelCount ; i > 0; i-- )
			{
				length = strlen(getText(i-1)) + 2; // 2 -> cr/lf after each text
				if (length > maxLength) 
				{
					maxLength = length;
				}
			}
		}
		
		// Count all additional exception location texts (if any)
		static const char* rethrowHeaderText = "\tfunction: %s, line: %d, file: %s\n";				
		if (exceptionLocationCount > 0)
		{
			for (int i = exceptionLocationCount ; i > 0; i-- )
			{
				length =	strlen(rethrowHeaderText) +
							strlen(name()) +
							strlen(exceptionLocation[i-1].getFunctionName()) +
							strlen(exceptionLocation[i-1].getFileName());
				if( length > maxLength) 
				{
					maxLength = length;
				}
			}
		}
	
		// Now allocate the buffer
		char * aWhatText = new char[maxLength + 1];
		
	
		// If the buffer is allocated, fill it with the message and use the trace function to print it.
		// By default the trace function prints to stderr
		if(aWhatText )
		{
			// Location of the exception
			if (exceptionLocationCount > 0)
			{
				sprintf(aWhatText ,
						throwHeaderText,
						name(),
						exceptionLocation[0].getFunctionName(),
						exceptionLocation[0].getLineNumber(),
						exceptionLocation[0].getFileName());
			}else
			{
				sprintf(aWhatText ,"\n%s exception thrown.\n   Location information is unavailable.\n", name());
			}
			whatText += std::string(aWhatText );
		}
		
		// ErrorID
		if (getErrorId())
		{
			sprintf(aWhatText ,"\tErrorId: %lu\n", getErrorId());
			whatText += std::string(aWhatText );
//			sprintf(aWhatText ,"\tError Message is %s", "todo"/*GetErrorIdAsString(errorId()).c_str()*/);
//			whatText += std::string(aWhatText );
		}else
		{
			sprintf(aWhatText ,"\tErrorId: unknown\n");
			whatText += std::string(aWhatText );
		}
		
		// As much info as we can get about the exception
		if (exceptionText)
		{
			sprintf(aWhatText ,"Exception text is:\n");
			whatText += std::string(aWhatText );
			for (int i = (getTextCount()-1); i >= 0 ;i-- )
			{
				sprintf(aWhatText ,"%s\n", getText(i));
				whatText += std::string(aWhatText );
			}
		}else
		{
			sprintf(aWhatText,"Exception text is unavailable.\n");
			whatText += std::string(aWhatText );
		}
		
		//	RETHROW information
		if (exceptionLocationCount > 0)
		{
			whatText += std::string("Rethrown at:\n");
			for (unsigned int i = 1; i < exceptionLocationCount ; i++ )
			{
				sprintf(	aWhatText,
							rethrowHeaderText,
							exceptionLocation[i].getFunctionName(),
							exceptionLocation[i].getLineNumber(),
							exceptionLocation[i].getFileName());
				whatText += std::string(aWhatText );
			}
		}		
		
		delete [] aWhatText ;
	
		// See the constructor of ExceptionText for implementation comments
		std::set_new_handler(oldHandler);
		
		return whatText.c_str();
	}
	
	 const char* Exception::name() const
	{
	   return "Exception";
	}
	
	bool Exception::isRecoverable() const
	{
		if(exceptionSeverity) return 1;
		return 0;
	}	

	/**
	 * Replacement for std::unexpected. For default behaviour: see 18.6.2.2.
	 * 
	 * Additional behaviour: adds tracing.
	 */
	/* static */ void Exception::unexpected()
	{
		Exception::getTraceFunction()->trace("Exception::unexpected() called.\n" );
		std::terminate();
	}
	
	/**
	 * Replacement for std::terminate. For default behaviour: see 18.6.3.2.
	 * 
	 * Additional behaviour: adds tracing.
	 */
	/* static */ void Exception::terminate()
	{
		Exception::getTraceFunction()->trace("Exception::terminate() called.\n" );
	  	std::abort();
	}
	/**
	 * The exception will be logged by sending the result of Exception.what() to 
	 * the Exception::getTraceFunction(). 
	 */
	void Exception::logException()
	{
		Exception::getTraceFunction()->trace(what());
	}		
		
	/**
	 * By default this function prints the text to stderr with the C function fprintf and
	 * does not do any formattting og the text (e.g append a cr/lf). 
	 * This function is NOT thread safe unless fprintf is.
	 * 
	 * If this is not the required behaviour, inherit your own class from Exception::TraceFunction
	 * and implement your own trace function (e.g tracing to a database or file). Use 
	 * Exception::setTraceFunction to replace the default TraceFunction. The Exception class assumes
	 * ownership of the pointer and will delete the pointer if the program terminates.
	 * */
	void Exception::TraceFunction::trace( const char* aText)
	{
	   	//fprintf(stderr,"%s", aText);
		OutputDebugString(aText);
	}
		
	// For an explanation see the hpp file
	DANU_EXCEPTION_CLASS_IMPLEMENT(AssertionFailure,Exception);
	DANU_EXCEPTION_CLASS_IMPLEMENT(InvalidParameter,Exception);
	DANU_EXCEPTION_CLASS_IMPLEMENT(InvalidRequest,Exception);
	DANU_EXCEPTION_CLASS_IMPLEMENT(ResourceExhausted,Exception);
	DANU_EXCEPTION_CLASS_IMPLEMENT(OutOfSystemResource,ResourceExhausted);
	DANU_EXCEPTION_CLASS_IMPLEMENT(OutOfWindowResource,ResourceExhausted);
	DANU_EXCEPTION_CLASS_IMPLEMENT(NotImplemented,Exception);
} // namespace Danu
