#ifndef Exception_hpp
#define Exception_hpp

#include "DanuConfig.h"
#include <stdexcept>

/// Default namespace for the library
namespace Danu
{
	class ExceptionText;

	/// Encapsulates the location where an exception is throw.
	class ExceptionLocation
	{
		public:
			ExceptionLocation(	const char* aFileName = 0,
								const char* aFunctionName = 0,
								unsigned long aLineNumber = 0 )
				:	fileName(aFileName),
					functionName(aFunctionName ? aFunctionName : "Unknown function"),
					lineNumber(aLineNumber)
			{}
			const char* getFileName() const
			{
				return fileName;
			}
			const char* getFunctionName() const
			{
				return functionName;
			}
			unsigned long getLineNumber() const
			{
				return lineNumber;
			}

		private:
			const char* fileName;
			const char* functionName;
			unsigned long lineNumber;
	}; // class ExceptionLocation

	class DANUIMPEXP Exception : public std::exception
	{
		public:
			enum Severity { unrecoverable, recoverable };

			class DANUIMPEXP TraceFunction
			{
				public:
					virtual void                        trace(	const char* aText);
			};	// class TraceFunction

			Exception(	const char* anExceptionText,
						unsigned long anErrorId = 0,
						Severity aSeverity = Exception::unrecoverable ) throw();
			Exception(	const std::string& anExceptionText,
						unsigned long anErrorId = 0,
						Severity aSeverity = Exception::unrecoverable ) throw();

			Exception(	const Exception& aException) throw();
			virtual ~Exception() throw();

			Severity						getSeverity() const;
			/// returns the platform dependant numeric value for the exception.
			unsigned long					getErrorId() const;
			/// returns the location where the exception is thrown
			const ExceptionLocation*		getLocation( unsigned long anIndex = 0) const;
			/// return the number of locations available where the excption is (re)thrown
			unsigned long					getLocationCount() const;
			/// returns the textual description of the reason the exception is thrown
			const char*						getText( unsigned long anIndex = 0 ) const;
			// return the number of textual description available for this exception
			unsigned long					getTextCount() const;

			void                         	setSeverity(Severity aSeverity);
			void                         	setErrorId( unsigned long anErrorId);
			virtual void					addLocation( const ExceptionLocation& aExceptionLocation);
			void                         	addText( const char* anErrorText) throw();
			virtual void					appendText( const char* anErrorText) throw();

			static void 						enableDefaultTrace();
			static void 						disableDefaultTrace();
			static void 						reEnableDefaultTrace();
			static void							setTraceFunction( Exception::TraceFunction* aTraceFunction);
			static Exception::TraceFunction*   	getTraceFunction();

			virtual const char*				what() const throw();

			virtual const char*				name() const;
			virtual bool					isRecoverable() const;

			static void						unexpected();
			static void 					terminate();

			virtual void					logException();

		private:
			/**
			 * The Trace class must be a friend so that it can ititialize
			 * the exceptionInitializer object
			 */
			friend class Trace;
			/// Class to initialize the exception environment
			class ExceptionInitializer
			{
				public:
					ExceptionInitializer();
					~ExceptionInitializer();

			}; // class ExceptionInitializer
			static ExceptionInitializer exceptionInitializer;

			Exception& operator=(const Exception& exc);

			Severity									exceptionSeverity;
			unsigned long								exceptionErrorId;
			ExceptionLocation							exceptionLocation[5];
			unsigned long								exceptionLocationCount;
			ExceptionText*								exceptionText;
			unsigned long								exceptionTextLevelCount;
			mutable std::string							whatText;
			static TraceFunction* 						defaultTraceFunction;
			static TraceFunction*						currentTraceFunction;

	}; // class Exception

	/**
	 * If Danuconfig.h is not included remove the DANU part of the macros
	 * and check if your compiler supports the __FUNCTION__ macro.
	 * The __FILE__ and __LINE__ macros are part of the ISO standard
	 */
	#define EXCEPTION_LOCATION()   ExceptionLocation(__DANUFILE__, __DANUFUNCTION__, __DANULINE__)

	/**
	 * Creates a declaration for a derived class of Exception or one of its derived classes.
	 * Parameters are the derived classname and it's parent classname
	*/
	#define DANU_EXCEPTION_CLASS_DECLARE(child,parent)				\
	class DANUIMPEXP child : public parent										\
	{																\
		public:														\
			child(	const char* a,									\
					unsigned long b = 0,							\
					Severity c = Exception::unrecoverable) throw();	\
			child(	const std::string& a,							\
					unsigned long b = 0,							\
					Severity c = Exception::unrecoverable) throw();	\
			child(const child &) throw();							\
			virtual ~child() throw();								\
			virtual const char* name() const;						\
		private:													\
			child &operator = ( const child & );					\
	}

	/**
	 * Creates the implementation for a derived class of Exception or one of its derived classes.
	 * Parameters are derived class name and it's parent class name
	 */
	#define DANU_EXCEPTION_CLASS_IMPLEMENT(child,parent)			\
	child::child(  const char* a,									\
	               unsigned long b,									\
	               Severity c) throw()								\
	   :  parent(a, b, c)											\
	{																\
	}																\
	child::child(const child & a) throw()							\
	   :  parent(a)													\
	{																\
	}																\
	child::~child() throw() 										\
	{																\
	}																\
	const char* child::name() const									\
	{																\
	   return ( # child);											\
	}
}

// Put these macros in global amespace to mimic the keywords
// It also seems that namespaces and macros do not like each other....

/**
 * Accepts as input a predefined object of any Exception-derived class.
 * The macro generates code to add the location information to the objects,
 * logs all object data, and throws the exception.
 */
#define THROW(exc)									\
	exc.addLocation(Danu::EXCEPTION_LOCATION()),	\
	exc.logException(),								\
	throw(exc)
/**
 * Accepts as input an object of any derived class of Exception that has been
 * previously thrown and caught. Like the THROW macro, it also captures the
 * location information and logs all object data before rethrowing the exception.
 */
#define RETHROW(exc)								\
	exc.addLocation(Danu::EXCEPTION_LOCATION()),	\
	exc.logException(),								\
	throw(exc)
/**
 * Creates and throws an exception of type type and with text text.
 */
#define THROW_EXCEPTION_OF_TYPE(type,text)			\
	type exc(text);									\
	THROW(exc);

	/*
	 * Creates and throws an exception of type type and with text text and sets the errorId
	 * of the exception to the result of GetErrorId() (On Windows that function calls GetLastError()).
	 *
	#define THROW_EXCEPTION_OF_TYPE_WITH_ERRORID(type,text)		type exc(text,GetErrorId());  \
																	THROW(exc);
	*/
namespace Danu
{
	/**
	 * The ASSERT macro expands to create and throw an object of the
	 * 	AssertionFailure class if the specified condition is not met.
	 */
	DANU_EXCEPTION_CLASS_DECLARE(AssertionFailure,Exception);
	#define THROW_ASSERTIONFAILURE(text) THROW_EXCEPTION_OF_TYPE(AssertionFailure,text)

	/**
	* When a member function detects an invalid input parameter, the member
	* function creates and throws an object of the InvalidParameter class.
	* This exception is identical to the exception AssertionFailure, with
	* one difference: InvalidParameter is thrown whether or not you
	* define DEVELOP for the compile.
	*/
	DANU_EXCEPTION_CLASS_DECLARE(InvalidParameter,Exception);
	#define THROW_INVALIDPARAMETER(text) THROW_EXCEPTION_OF_TYPE(InvalidParameter,text)
	/**
	 * Whenever an object cannot satisfy a request, the member function creates
	 * and throws an object of the InvalidRequest class. An example of such a
	 * request occurs if you try to get a not existing object in a container
	 */
	DANU_EXCEPTION_CLASS_DECLARE(InvalidRequest,Exception);
	#define THROW_INVALIDREQUEST(text) THROW_EXCEPTION_OF_TYPE(InvalidRequest,text)

	/**
	 * When a member function makes a resource request of the operating system
	 * or the presentation system that it cannot satisfy, the member function
	 * creates and throws an object of the ResourceExhausted class or one of
	 * its derived classes. ResourceExhausted is the generic out-of-resource class.
	 * Member functions use ResourceExhausted whenever its derived classes, which
	 * are for specific out-of-resource cases, are not applicable.
	 */
	DANU_EXCEPTION_CLASS_DECLARE(ResourceExhausted,Exception);
	#define THROW_RESOURCEEXHAUSTED(text) THROW_EXCEPTION_OF_TYPE(ResourceExhausted,text)

	/**
	 * When a member function makes an operating system resource request
	 * that the system cannot satisfy, the member function creates and
	 * throws an object of the OutOfSystemResource class.
	 */
	DANU_EXCEPTION_CLASS_DECLARE(OutOfSystemResource,ResourceExhausted);
	#define THROW_OUTOFSYSTEMRESOURCES(text) THROW_EXCEPTION_OF_TYPE(OutOfSystemResource,text)

	/**
	 * When a member function makes a presentation (window) system resource request
	 * that the system cannot satisfy, the member function creates and throws
	 * an object of the OutOfWindowResource class.
	 */
	DANU_EXCEPTION_CLASS_DECLARE(OutOfWindowResource,ResourceExhausted);
	#define THROW_OUTOFWINDOWSRESOURCES(text) THROW_EXCEPTION_OF_TYPE(OutOfWindowResource,text)

	/**
	 * Whenever a function is called that is not implemented and it should
	 * be: a replacement for pure virtual functions during development.
	 */
	DANU_EXCEPTION_CLASS_DECLARE(NotImplemented,Exception);
	#define THROW_NOTIMPLEMENTED(text) THROW_EXCEPTION_OF_TYPE(NotImplemented,text)
} // namespace Danu

#endif //_Exception_hpp_
