#include "TextProfile.hpp"

namespace Danu
{

	// Default ctor is needed for the pair<> template expansion....
	SectionItem::SectionItem()
	   :  name(""),
	      value(""),
	      comment("")
	{
	}

	SectionItem::SectionItem(const char* aNameValueAndComment)
	{
	   splitString(   aNameValueAndComment,
	                  name,
	                  '=',
	                  value,
	                  ';',
	                  comment);

	   stripLeadingAndTrailingSpaces(  name);
	   stripLeadingAndTrailingSpaces(  value);
	   stripLeadingAndTrailingSpaces(  comment);
	}

	SectionItem::SectionItem(const char* aName,
	                           const char* aValue,
	                           const char* aComment /* = 0 */)
	   :  name(aName),
	      value(aValue),
	      comment(aComment ? aComment : "")
	{
	}

	SectionItem::SectionItem(const char* aName,
	                           const long aValue,
	                           const char* aComment /* = 0 */)
	   :  name(aName),
	      value(integerToString(aValue)),
	      comment(aComment ? aComment : "")
	{
	}

	SectionItem::SectionItem(const SectionItem& SectionItem)
	   :  name(SectionItem.name),
	      value(SectionItem.value),
	      comment(SectionItem.comment)
	{
	   if (this != &SectionItem)
	   {
	   }
	}

	// Destructor
	SectionItem::~SectionItem()
	{
	}

	// Operators
	SectionItem& SectionItem::operator=(const SectionItem& SectionItem)
	{
	   if (this != &SectionItem)
	   {
	      name     = SectionItem.name;
	      value    = SectionItem.value;
	      comment  = SectionItem.comment;
	   }
	   return *this;
	}

	bool SectionItem::operator==(const SectionItem& SectionItem) const
	{
	   return (name == SectionItem.name) && (value == SectionItem.value) && (comment == SectionItem.comment);
	}

	bool SectionItem::operator<(const SectionItem& SectionItem) const
	{
	   if(name <  SectionItem.name) return true;
	   if( (name == SectionItem.name) && (value < SectionItem.value)) return true;
	   return false;
	}

	SectionItem::operator std::string& () const
	{
	   return (std::string&)value;
	}

	SectionItem::operator long () const
	{
	   return stringToLong((std::string&)value);
	}

	// Accessors
	std::string& SectionItem::getName() const
	{
	   return (std::string&)name;
	}

	std::string& SectionItem::getValueAsString() const
	{
	   return (std::string&)value;
	}

	long SectionItem::getValueAsLong() const
	{
	   return stringToLong((std::string&)value);
	}

	std::string& SectionItem::getComment() const
	{
	   return (std::string&)comment;
	}

	// Mutators
	void SectionItem::setName(const char* aName)
	{
	   name = aName;
	}

	void SectionItem::setValue(  const char* aValue,
	                              const char* aComment /* = 0 */)
	{
	   value = aValue;
	   if(aComment)
	   {
	      comment = aComment;
	   }
	}

	void SectionItem::setValue(  const long aValue,
	                              const char* aComment /* = 0 */)
	{
	   value = integerToString(aValue);
	   if(aComment)
	   {
	      comment = aComment;
	   }
	}

	void SectionItem::setComment(const char* aComment)
	{
	   comment = aComment;
	}

	// Other operations

	// protected
	// Operations

	// private
	// Operations

	// End of class definition of SectionItem

	// Friend output operator
	// This operator will put an endl into the stream as a profile is line oriented
	DANUIMPEXP std::ostream& operator<<(std::ostream& os,const SectionItem& SectionItem)
	{
	   os << SectionItem.name << " = " << SectionItem.value;
	   if (SectionItem.comment != std::string(""))
	   {
	      os << "   ; " << SectionItem.comment;
	   }
	   os << std::endl;
	   return os;
	}

	//******************************************* Section *******************************************
	// public
	// Constructors
	// Default ctor is needed for the pair<> template expansion....
	Section::Section()
	   :  name(""),
	      comment("")
	{
	}

	Section::Section(  const char* SectionNameWithComment)
	{
	   splitString(   SectionNameWithComment,
	                  name,
	                  ';',
	                  comment);

	   stripLeadingAndTrailingSpaces(  name);
	   name = bracketize(name);

	   stripLeadingAndTrailingSpaces(  comment);
	}

	Section::Section(  const char* SectionName,
	                     const char* aComment)
	   :  name( SectionName),
	      comment( aComment ? aComment : "")
	{
	   name = bracketize(name);
	}

	Section::Section(const Section& aSection)
	   :  name( aSection.name),
	      comment( aSection.comment),
	      sectionItems(aSection.sectionItems)
	{
	   if (this != &aSection)
	   {
	   }
	}

	// Destructor
	Section::~Section()
	{
	}

	// Operators
	Section& Section::operator=(const Section& aSection)
	{
	   if (this != &aSection)
	   {
	      name           = aSection.name;
	      comment        = aSection.comment;
	      sectionItems   = aSection.sectionItems;
	   }
	   return *this;
	}

	bool Section::operator==(const Section& aSection) const
	{
	   return this == &aSection;
	}

	bool Section::operator<(const Section& aSection) const
	{
	   return this < &aSection;
	}

	// Accessors
	std::string& Section::getName() const
	{
	   return (std::string&)name;
	}

	std::string& Section::getComment() const
	{
	   return (std::string&)comment;
	}

	// Mutators
	void Section::setName(const char* aName)
	{
	   name = bracketize(aName);
	}

	void Section::setComment(const char* aComment)
	{
	   comment = aComment;
	}

	// Other operations
	unsigned long Section::getNumberOfSectionItems() const
	{
	   return sectionItems.size();
	}

	// Testing for presence
	bool Section::containsSectionItem( const char* SectionItemName) const
	{
	   std::map< std::string, SectionItem, std::less<std::string> >::const_iterator i  = sectionItems.find(SectionItemName);
	   if(i==sectionItems.end())
	   {
	      return false;
	   }else
	   {
	      return true;
	   }
	}

	// Getting it
	SectionItem& Section::getSectionItem(const char* SectionItemName) const
	{
	   std::map< std::string, SectionItem, std::less<std::string> >::const_iterator i  = sectionItems.find(SectionItemName);
	   if(i==sectionItems.end())
	   {
	      std::string msg = std::string("\tSection: ") + getName() + std::string("\n\tSectionItem ") + std::string(SectionItemName) + std::string(" not found");
	      Danu::InvalidRequest exc(msg.c_str());
	      THROW(exc);
	   }else
	   {
	      return (SectionItem&)(*i).second;
	   }
	}

	std::string& Section::getSectionItemAsString( const char* SectionItemName) const
	{
	   return getSectionItem(SectionItemName).getValueAsString();
	}

	long Section::getSectionItemAsLong(const char* SectionItemName) const
	{
	   return getSectionItem(SectionItemName).getValueAsLong();
	}

	// Adding one. If already present exception is thrown.
	void Section::addSectionItem(const SectionItem& aSectionItem)
	{
	   // Inserts t into a if and only if a does not already contain
	   // an element whose key is the same as the key of t.

	   // The return value is a pair P.
	   // P.first is an iterator pointing to the element whose key is the same as the key of t.
	   // P.second is a bool: it is true if t was actually inserted into a, and false
	   // if t was not inserted into a, i.e. if a already contained an element with the same key as t.

	   // P.first is a dereferenceable iterator.
	   // *(P.first) has the same key as t.
	   // The size of a is incremented by 1 if and only if P.second is true.

	   std::pair< std::map< std::string,Danu::SectionItem,std::less<std::string> >::iterator, bool> result = sectionItems.insert( std::pair< const std::string, Danu::SectionItem > (aSectionItem.getName(),aSectionItem) );
	   if(result.second == false)
	   {
	      // Already contained so throw the exception
	      std::string msg = std::string("\tSection: ") + getName() + std::string("\n\tSectionItem ") + aSectionItem.getName() + std::string(" already in profile");
	      Danu::InvalidRequest exc(msg.c_str());
	      THROW(exc);
	   }
	}

	void Section::addSectionItem(const char* SectionItemName,
	                              const char* SectionItemValue,
	                              const char* SectionItemComment /* = 0 */)
	{
	   addSectionItem(SectionItem(SectionItemName,SectionItemValue,SectionItemComment));
	}

	void Section::addSectionItem(const char* SectionItemName,
	                              const long SectionItemValue,
	                              const char* SectionItemComment /* = 0 */)
	{
	   addSectionItem(SectionItem(SectionItemName,SectionItemValue,SectionItemComment));
	}

	// Adding one. If already present replace the existing one with the new one.
	void Section::addOrReplaceSectionItem(const SectionItem& aSectionItem)
	{
	   // Inserts t into a if and only if a does not already contain
	   // an element whose key is the same as the key of t.

	   // The return value is a pair P.
	   // P.first is an iterator pointing to the element whose key is the same as the key of t.
	   // P.second is a bool: it is true if t was actually inserted into a, and false
	   // if t was not inserted into a, i.e. if a already contained an element with the same key as t.

	   // P.first is a dereferenceable iterator.
	   // *(P.first) has the same key as t.
	   // The size of a is incremented by 1 if and only if P.second is true.

	   std::pair< std::map< std::string,Danu::SectionItem,std::less<std::string> >::iterator, bool> result = sectionItems.insert( std::pair< const std::string, Danu::SectionItem > (aSectionItem.getName(),aSectionItem) );
	   if(result.second == false)
	   {
	      // Already contained so replace
	      (*result.first).second = aSectionItem;
	   }
	}

	void Section::addOrReplaceSectionItem(const char* aSectionItemName,
	                                       const char* aSectionItemValue,
	                                       const char* aSectionItemComment /* = 0 */)
	{
	   addOrReplaceSectionItem(SectionItem(aSectionItemName,aSectionItemValue,aSectionItemComment));
	}

	void Section::addOrReplaceSectionItem(const char* aSectionItemName,
	                                       const long aSectionItemValue,
	                                       const char* aSectionItemComment /* = 0 */)
	{
	   addOrReplaceSectionItem(SectionItem(aSectionItemName,aSectionItemValue,aSectionItemComment));
	}

	// Removing one.
	void Section::removeSectionItem( const SectionItem& SectionItem)
	{
	   removeSectionItem( SectionItem.getName().c_str());
	}

	void Section::removeSectionItem(  const char* SectionItemName)
	{
	   std::map< std::string, SectionItem, std::less<std::string> >::iterator i  = sectionItems.find(SectionItemName);
	   if(i==sectionItems.end())
	   {
	      std::string msg = std::string("\tSection: ") + getName() + std::string("\n\tSectionItem ") + std::string(SectionItemName) + std::string(" not found");
	      Danu::InvalidRequest exc(msg.c_str());
	      THROW(exc);
	   }else
	   {
	   }
	}

	// Iterator stuff
	std::map< std::string, SectionItem ,std::less<std::string> >::iterator Section::sectionItemsBegin()
	{
	   return sectionItems.begin();
	}

	std::map< std::string, SectionItem ,std::less<std::string> >::const_iterator Section::sectionItemsBegin() const
	{
	   return sectionItems.begin();
	}

	std::map< std::string, SectionItem ,std::less<std::string> >::iterator Section::sectionItemsEnd()
	{
	   return sectionItems.end();
	}

	std::map< std::string, SectionItem ,std::less<std::string> >::const_iterator Section::sectionItemsEnd() const
	{
	   return sectionItems.end();
	}

	// protected
	// Operations

	// private
	// Operations

	// End of class definition of Section

	// Friend output operator
	// This operator will put an endl into the stream as a profile is line oriented
	DANUIMPEXP std::ostream& operator<<(std::ostream& os,const Section& aSection)
	{
	   os << aSection.name;
	   if (aSection.comment != std::string(""))
	   {
	      os << "   ; "<< aSection.comment;
	   }
	   os << std::endl;

	   for(  std::map< std::string, SectionItem, std::less<std::string> >::const_iterator i  = aSection.sectionItems.begin();
	         i != aSection.sectionItems.end();
	         i++)
	   {
	      os << (*i).second;
	   }
	   // Put two endls into the stream so that sections will be seperated by an empty line
	   os << std::endl << std::endl;
	   return os;
	}

	//******************************************* TextProfile *******************************************
	// public
	// Constructors
	TextProfile::TextProfile()
	   :  fileName(""),
	      defaultSection("")
	{
	}

	TextProfile::TextProfile(const char* aFileName,
	                           const char* aDefaultSectionName /* = 0 */)
	   :  fileName(aFileName),
	      defaultSection("")
	{
	   if(aDefaultSectionName)
	   {
	      defaultSection = bracketize(aDefaultSectionName);
	   }

	   // All the stuff will be in memory
	   // Gives good performance but bad memory use
	   readFile();
	}

	TextProfile::TextProfile(std::string aFileName,
	                           const char* aDefaultSectionName /* = 0 */)
	   :  fileName(aFileName),
	      defaultSection("")
	{
	   if(aDefaultSectionName)
	   {
	      defaultSection = bracketize(aDefaultSectionName);
	   }

	   // All the stuff will be in memory
	   // Gives good performance but bad memory use
	   readFile();

	}

	TextProfile::TextProfile(const TextProfile& aTextProfile)
	   :  fileName(aTextProfile.fileName),
	      defaultSection(aTextProfile.defaultSection)
	{
	   // Let it be noted that not the memory image but the diskimage is copied!!!!!
	   readFile();
	}

	// Destructor
	TextProfile::~TextProfile()
	{
	   writeFile();

	   sections.erase(sections.begin(),sections.end());
	}

	// Operators
	TextProfile& TextProfile::operator=(const TextProfile& aTextProfile)
	{
	   if (this != &aTextProfile)
	   {
	      fileName       = aTextProfile.fileName;
	      defaultSection = aTextProfile.defaultSection;

	      // Delete all old stuff
	      sections.erase(sections.begin(),sections.end());

	      // Refresh the contents
	      // Let it be noted that not the memory image but the diskimage is copied!!!!!
	      readFile();
	   }
	   return *this;
	}

	bool TextProfile::operator==(const TextProfile& aTextProfile) const
	{
	   return fileName == aTextProfile.fileName;
	}

	bool TextProfile::operator<(const TextProfile& aTextProfile) const
	{
	   return fileName < aTextProfile.fileName;
	}

	// Accessors
	std::string& TextProfile::getName() const
	{
	   return (std::string&)fileName;
	}

	std::string& TextProfile::getDefaultSectionName() const
	{
	   return (std::string&)defaultSection;
	}

	std::string& TextProfile::getComment() const
	{
	   return (std::string&)comment;
	}

	// Mutators
	void TextProfile::setName( const char* aFileName)
	{
	   fileName = aFileName;
	}

	void TextProfile::setDefaultSectionName(  const char* SectionName)
	{
	   // Let the user give a name without brackets
	   defaultSection = bracketize(SectionName);
	   // If the default section does not exit, insert an empty one!!!
	   if(!containsSection(SectionName))
	   {
	      addSection(SectionName);
	   }
	}

	void TextProfile::setComment( const char* aComment)
	{
	   comment = aComment;
	}

	// Other operations
	unsigned long TextProfile::getNumberOfSections() const
	{
	   return sections.size();
	}

	// Testing for presence
	bool TextProfile::containsSection( const char* aSectionName) const
	{
	   // Check if this is a valid Section
	   //ASSERT(Danu::InvalidParameter,aSectionName != 0);

	   // Let the user give a name without brackets
	   std::string sectionName = bracketize(aSectionName);

	   std::map< std::string, Section ,std::less<std::string> >::const_iterator i = sections.find(sectionName.c_str());
	   if(i==sections.end())
	   {
	      return false;
	   }else
	   {
	      return true;
	   }
	}

	// Getting it
	Section& TextProfile::getSection( const char* aSectionName) const
	{
	   // Check if this is a valid Section
	   //ASSERT(Danu::InvalidParameter,strlen(aSectionName) != 0);

	   // Let the user give a name without brackets
	   std::string sectionName = bracketize(aSectionName);

	   //ATRACE_DEVELOP(strprintf("Trying to retrieve section %s",sectionName.c_str()));

	   std::map< std::string, Section ,std::less<std::string> >::const_iterator i = sections.find(sectionName.c_str());
	   if(i==sections.end())
	   {
	      //ATRACE_DEVELOP(strprintf("Section %s NOT found",sectionName.c_str()));
	      std::string msg = std::string("\tTextProfile: ") + getName() + std::string("\n\tSection: ") + sectionName + std::string(" not found");
	      Danu::InvalidRequest exc(msg.c_str());
	      THROW(exc);
	   }else
	   {
	      //ATRACE_DEVELOP(strprintf("Section %s found",sectionName.c_str()));
	      return (Section&)(*i).second;
	   }
	}

	// Adding one. If already present exception is thrown.
	void TextProfile::addSection( const Section& aSection)
	{
	   // Inserts t into a if and only if a does not already contain
	   // an element whose key is the same as the key of t.

	   // The return value is a pair P.
	   // P.first is an iterator pointing to the element whose key is the same as the key of t.
	   // P.second is a bool: it is true if t was actually inserted into a, and false
	   // if t was not inserted into a, i.e. if a already contained an element with the same key as t.

	   // P.first is a dereferenceable iterator.
	   // *(P.first) has the same key as t.
	   // The size of a is incremented by 1 if and only if P.second is true.

	   std::pair< std::map< std::string,Section,std::less<std::string> >::iterator, bool> result = sections.insert( std::pair< const std::string, Section > (aSection.getName(),aSection) );
	   if(result.second == false)
	   {
	      std::string msg = std::string("\tTextProfile: ") + getName() + std::string("\n\tSection: ") + aSection.getName() + std::string(" already in profile");
	      Danu::InvalidRequest exc(msg.c_str());
	      THROW(exc);
	//*******************************************************************
	   }
	}

	void TextProfile::addSection(const char* aSectionName,
	                              const char* aSectionComment /* = 0 */)
	{
	   addSection(Section(aSectionName,aSectionComment));
	}

	// Adding one. If allready present replace the existing one with the new one.
	void TextProfile::addOrReplaceSection(const Section& aSection)
	{
	   // Inserts t into a if and only if a does not already contain
	   // an element whose key is the same as the key of t.

	   // The return value is a pair P.
	   // P.first is an iterator pointing to the element whose key is the same as the key of t.
	   // P.second is a bool: it is true if t was actually inserted into a, and false
	   // if t was not inserted into a, i.e. if a already contained an element with the same key as t.

	   // P.first is a dereferenceable iterator.
	   // *(P.first) has the same key as t.
	   // The size of a is incremented by 1 if and only if P.second is true.

	   std::pair< std::map< std::string,Section,std::less<std::string> >::iterator, bool> result = sections.insert( std::pair< const std::string, Section > (aSection.getName(),aSection) );
	   if(result.second == false)
	   {
	      // Allready contained so replace
	      (*result.first).second = aSection;
	   }
	}

	void TextProfile::addOrReplaceSection(const char* aSectionName,
	                                       const char* aSectionComment /* = 0 */)
	{
	   addOrReplaceSection(Section(aSectionName,aSectionComment));
	}

	// Removing one
	void TextProfile::removeSection( const Section& aSection)
	{
	}

	void TextProfile::removeSection( const char* aSectionName)
	{
	   removeSection(Section(aSectionName,0));
	}


	// SectionItem stuff: if aSectionName == 0 the default Section will be used.
	// If no default section is available an exception is thrown.
	// Testing for presence
	bool TextProfile::containsSectionItem( const char* SectionItemName,
	                                        const char* SectionName /* = 0 */) const
	{
	   if(containsSection(SectionName ? SectionName : defaultSection.c_str()))
	   {
	      return getSection(SectionName ? SectionName : defaultSection.c_str()).containsSectionItem(SectionItemName);
	   }
	   return false;
	}

	// Getting it
	SectionItem& TextProfile::getSectionItem(  const char* aSectionItemName,
	                                             const char* aSectionName /* = 0 */) const
	{
	   return getSection(aSectionName ? aSectionName : defaultSection.c_str()).getSectionItem(aSectionItemName);
	}

	std::string& TextProfile::getSectionItemAsString(	const char* aSectionItemName,
	                                         		const char* aSectionName /* = 0 */) const
	{
	   return getSectionItem(aSectionItemName,aSectionName).getValueAsString();
	}

	long TextProfile::getSectionItemAsLong(  const char* aSectionItemName,
	                                          const char* aSectionName /* = 0 */) const
	{
	   return getSectionItem(aSectionItemName,aSectionName).getValueAsLong();
	}

	// Adding one. If already present exception is thrown.
	void TextProfile::addSectionItem(  const SectionItem& aSectionItem,
	                                    const char* aSectionName /* = 0 */)
	{
	   getSection( aSectionName ? aSectionName : defaultSection.c_str()).addSectionItem(aSectionItem);
	}

	void TextProfile::addSectionItem(  const char* aSectionItemName,
	                                    const char* aSectionItemValue,
	                                    const char* aSectionItemComment /* = 0 */,
	                                    const char* aSectionName /* = 0 */)
	{
	   addSectionItem(SectionItem(aSectionItemName,aSectionItemValue,aSectionItemComment),aSectionName);
	}

	void TextProfile::addSectionItem(  const char* aSectionItemName,
	                                    const long aSectionItemValue,
	                                    const char* aSectionItemComment /* = 0 */,
	                                    const char* aSectionName /* = 0 */)
	{
	   addSectionItem(SectionItem(aSectionItemName,aSectionItemValue,aSectionItemComment),aSectionName);
	}

	// Adding or replacing (setting) one.
	void TextProfile::addOrReplaceSectionItem(  const SectionItem& aSectionItem,
	                                             const char* aSectionName /* = 0 */)
	{
	   // This function assumes that the SectionName or defaultSection exists...
	   getSection(aSectionName ? aSectionName : defaultSection.c_str()).addOrReplaceSectionItem(aSectionItem);
	}

	void TextProfile::addOrReplaceSectionItem(  const char* aSectionItemName,
	                                             const char* aSectionItemValue,
	                                             const char* aSectionItemComment /* = 0 */,
	                                             const char* aSectionName /* = 0 */ )
	{
	   addOrReplaceSectionItem(SectionItem(aSectionItemName,aSectionItemValue,aSectionItemComment),aSectionName);
	}

	void TextProfile::addOrReplaceSectionItem(  const char* aSectionItemName,
	                                             const long aSectionItemValue,
	                                             const char* aSectionItemComment /* = 0 */,
	                                             const char* aSectionName /* = 0 */)
	{
	   addOrReplaceSectionItem(SectionItem(aSectionItemName,aSectionItemValue,aSectionItemComment),aSectionName);
	}

	// Removing one
	void TextProfile::removeSectionItem(  const SectionItem& aSectionItem,
	                                       const char* aSectionName /* = 0 */)
	{
	   // Development tracing
	//   AFUNCTRACE_DEVELOP();

	   getSection(aSectionName ? aSectionName : defaultSection.c_str()).removeSectionItem(aSectionItem);
	}

	void TextProfile::removeSectionItem(  const char* aSectionItemName,
	                                       const char* aSectionName /* = 0 */)
	{
	   removeSectionItem(SectionItem(aSectionItemName,(char*)0,0),aSectionName);
	}

	// Misc functions
	// Rereading the profile from disk
	void TextProfile::refresh()
	{
	   // Delete all in memory
	   sections.erase(sections.begin(),sections.end());

	   // Read the file again
	   readFile();
	}

	void TextProfile::save()
	{
	   // Just rewrite the file
	   writeFile();
	}

	// Iterator stuff
	std::map< std::string, Section ,std::less<std::string> >::iterator TextProfile::sectionsBegin()
	{
	   return sections.begin();
	}

	std::map< std::string, Section ,std::less<std::string> >::const_iterator TextProfile::sectionsBegin() const
	{
	   return sections.begin();
	}

	std::map< std::string, Section ,std::less<std::string> >::iterator TextProfile::sectionsEnd()
	{
	   return sections.end();
	}

	std::map< std::string, Section ,std::less<std::string> >::const_iterator TextProfile::sectionsEnd() const
	{
	   return sections.end();
	}

	// protected
	// Operations
	void TextProfile::readFile()
	{
	   // Check if the file exists, if not create an empty profile
	   // the parsing failes on an empty file. Don't know why
	   // access is POSIX so portable???

	   //  00      Check only for the existence of the file.
	   //  02      Check for permission to write to the file.
	   //  04      Check for permission to read from the file.
	   //  06      Check for permission to read from and write to the file.

	   // Return Value
	   // access returns 0 if you can get access to the file in the specified mode.
	   //  A return value of -1 shows that the file does not exist or is inaccessible
	   // in the given mode, and the system sets errno to one of the following values:


	   // Value          Meaning
	   // EACCESS        Access is denied; the permission setting of the file does not allow you to get access to the file in the
	   //                specified mode.
	   // ENOENT         The system cannot find the file or the path that you specified, or the file name was incorrect.
	   // EINVAL         The mode specified was not valid.
	   // EOS2ERR        The call to the operating system was not successful.
	/*
	   if (access(fileName.c_str(),00)==0)
	   {
	//      ATRACE_DEVELOP("File does not exist");
	      ifstream file(fileName.c_str());
	   }
	   if (access(fileName.c_str(),02)==0)
	   {
	//      ATRACE_DEVELOP("File cannot be written to");
	      return;
	   }
	   if (access(fileName.c_str(),04)==0)
	   {
	//      ATRACE_DEVELOP("File cannot be read from");
	      return;
	   }
	   if (access(fileName.c_str(),06)==-1)
	   {
	//      ATRACE_DEVELOP("File cannot be read from or written to");
	      return;
	   }
	*/
	   std::ifstream file(fileName.c_str());

	   char lineBuffer[256] = {0};

	   // While parsing the profile the following is assumed:
	   // 1. It starts with an optional comment, no empty line is allowed !!!!
	   // 2. After the comment a section starts
	   // 3. No comments between the sections
	   // 4. In a section, no empty lines are allowed and a value may not begin with '['

	   // Check for an empty file. If empty skip the reading of the file
	   if (file.peek() == EOF)
	   {
		  file.close();
	      return;
	   }

	   // Look for comment. In case of an empty profile file this will be skipped
	   while(file.peek() == ';')
	   {
	      file.getline(&lineBuffer[0],255);
	      comment = comment + std::string(lineBuffer);
	      comment = comment + std::string("\r\n");
	   }

	   // Skip any blank lines untill a start section delimiter is encountered.
	   // Exit in case the profile file consists only of a introductory comment
	   while ( file.peek() != '[')
	   {
	      int c = file.get();
	      if (c == EOF)
	      {
	         // 18-9-99: it seems that eof() is not set???? It is not reproducable.... So comment out for now.
	         /*
	         if (file.eof() == false)
	         {
	            cout << "EOF encountered but eof() still returns false" << endl;
	         }
	         */
	         break;
	      }
	   }

	   while (file.eof() == false)
	   {
	      // Read the section header
	      file.getline(&lineBuffer[0],255);
	//      cout << "Found section: " << lineBuffer << endl;
	      Section section(&lineBuffer[0]);

	      // Read the SectionItems until a line beginning with '[' is encountered
	      // Skip any blank lines
	      while (file.peek() != '[')
	      {
	         file.getline(&lineBuffer[0],255);
	         std::string line(&lineBuffer[0]);
	         if(line.length() > 0)
	         {
	//            cout << "Found sectionItem: " << lineBuffer << endl;
	            SectionItem sectionItem(&lineBuffer[0]);
	            section.addOrReplaceSectionItem(sectionItem);
	         }else
	         {
	//            cout << "Found empty line: " << lineBuffer << endl;
	         }
	         if( file.eof() == true) break;
	      }
	      std::pair< std::map< std::string,Section,std::less<std::string> >::iterator, bool> result = sections.insert( std::pair< const std::string, Section > (section.getName(),section) );
	      if(result.second == false)
	      {
	         std::string exceptionText("Insertion of aSection failed");
	         Danu::InvalidRequest exc(exceptionText.c_str());
	         THROW(exc);
	      }
	   }

	   file.close();
	}

	void TextProfile::writeFile()
	{
	   std::ofstream file(fileName.c_str());
	   file << *this;
	   file.close();
	}

	// private
	// Operations

	// End of class definition of TextProfile

	// Friend output operator
	// This operator will put an endl into the stream as a profile is line oriented
	DANUIMPEXP std::ostream& operator<<(std::ostream& os,const TextProfile& aTextProfile)
	{
	   os << aTextProfile.comment << std::endl;

	   for(  std::map< std::string, Section, std::less<std::string> >::const_iterator i  = aTextProfile.sections.begin();
	         i != aTextProfile.sections.end();
	         i++)
	   {
	      os << (*i).second;
	   }
	   os << std::endl;
	   return os;
	}

} // namespace Danu
