#ifndef TextProfile_hpp
#define TextProfile_hpp

#include "DanuConfig.h"

#include <iostream>
#include <fstream>
#include <map>

#include "DanuString.hpp"

namespace Danu
{
	class DANUIMPEXP SectionItem
	{
		public:
			// Default ctor is needed for the pair<> template expansion....
			SectionItem();
			// aNameValueAndComment will be parsed into their parts
			// This is handy if an item is read from file
			SectionItem(const char* aNameValueAndComment);
			SectionItem(const char* aName,
						const char* aValue,
						const char* aComment = 0);
			SectionItem(const char* aName,
						const long aValue,
						const char* aComment = 0);
			SectionItem(const SectionItem& SectionItem);

			virtual ~SectionItem();

			SectionItem& operator=(const SectionItem& SectionItem);
			bool operator==(const SectionItem& SectionItem) const;
			bool operator<(const SectionItem& SectionItem) const;

			operator std::string& () const;
			operator long () const;


			std::string&                              getName() const;
			std::string&                              getValueAsString() const;
			long                                      getValueAsLong() const;
			std::string&                              getComment() const;


			void                                      setName(	const char* aName);
			void                                      setValue(	const char* aValue,
																const char* aComment = 0);
			void                                      setValue(	const long aValue,
																const char* aComment = 0);
			void                                      setComment( const char* aComment);

	   protected:
	      std::string                                    name;
	      std::string                                    value;
	      std::string                                    comment;

	   private:
	      // This operator will put an endl into the stream as a profile is line oriented
	      friend DANUIMPEXP std::ostream& operator<<(std::ostream& os,const SectionItem& SectionItem);

	}; // class SectionItem

	class DANUIMPEXP Section
	{
		public:
			// Defalt ctor is needed for the pair<> template expansion....
			Section();
			// SectionNameWithComment will be parsed into their parts
			// This is handy if a section is read from file
			Section(const char* SectionNameWithComment);
			Section(const char* SectionName,
					const char* aComment);
			Section(const Section& aSection);

			virtual ~Section();

			Section& operator=(const Section& aSection);
			bool operator==(const Section& aSection) const;
			bool operator<(const Section& aSection) const;

			std::string&                              getName() const;
			std::string&                              getComment() const;

			void                                      setName( const char* aName);
			void                                      setComment(const char* aComment);

			unsigned long                             getNumberOfSectionItems() const;
			// Testing for presence
			bool                                      containsSectionItem( const char* SectionItemName) const;
			// Getting it
			SectionItem&                              getSectionItem(const char* SectionItemName) const;
			std::string&                              getSectionItemAsString( const char* SectionItemName) const;
			long                                      getSectionItemAsLong(const char* SectionItemName) const;
			// Adding one. If allready present exception is thrown.
			void                                      addSectionItem(	const SectionItem& SectionItem);
			void                                      addSectionItem(	const char* SectionItemName,
																		const char* SectionItemValue,
																		const char* SectionItemComment = 0);
			void                                      addSectionItem(	const char* SectionItemName,
																		const long SectionItemValue,
																		const char* SectionItemComment = 0);
			// Adding one. If allready present replace the existing one with the new one.
			void                                      addOrReplaceSectionItem(	const SectionItem& SectionItem);
			void                                      addOrReplaceSectionItem(	const char* SectionItemName,
																				const char* SectionItemValue,
																				const char* SectionItemComment = 0);
			void                                      addOrReplaceSectionItem(	const char* SectionItemName,
																				const long SectionItemValue,
																				const char* SectionItemComment = 0);
			// Removing one.
			void                                      removeSectionItem(const SectionItem& SectionItem);
			void                                      removeSectionItem(const char* SectionItemName);

			// Iterator stuff. Should be typedef-ed ????
			std::map< std::string, SectionItem ,std::less<std::string> >::iterator             sectionItemsBegin();
			std::map< std::string, SectionItem ,std::less<std::string> >::const_iterator       sectionItemsBegin() const;
			std::map< std::string, SectionItem ,std::less<std::string> >::iterator             sectionItemsEnd();
			std::map< std::string, SectionItem ,std::less<std::string> >::const_iterator       sectionItemsEnd() const;

		protected:
			std::string                                    name;
			std::string                                    comment;
			std::map< std::string, SectionItem ,std::less<std::string> > sectionItems;

	   private:
			// Friends
			// This operator will put an endl into the stream as a profile is line oriented
			friend DANUIMPEXP std::ostream& operator<<(std::ostream& os,const Section& aSection);
	}; // class Section

	class DANUIMPEXP TextProfile
	{
		public:
			TextProfile();
			TextProfile(const char* aFileName,
						const char* aDefaultSectionName = 0);
			TextProfile(std::string aFileName,
						const char* aDefaultSectionName = 0);
			TextProfile(const TextProfile& aTextProfile);

			virtual ~TextProfile();

			TextProfile& operator=(const TextProfile& aTextProfile);
			bool operator==(const TextProfile& aTextProfile) const;
			bool operator<(const TextProfile& aTextProfile) const;

			std::string&                              getName() const;
			std::string&                              getDefaultSectionName() const;
			std::string&                              getComment() const;

			void                                      setName( const char* aFileName);
			void                                      setDefaultSectionName(  const char* SectionName);
			void                                      setComment( const char* aComment);

			unsigned long                             getNumberOfSections() const;
			bool                                      containsSection( const char* SectionName) const;
			Section&                                  getSection( const char* SectionName) const;
			// Adding one. If allready present exception is thrown.
			void                                      addSection(	const Section& aSection);
			void                                      addSection(	const char* SectionName,
																	const char* SectionComment = 0);
			// Adding one. If allready present replace the existing one with the new one.
			void                                      addOrReplaceSection( const Section& aSection);
			void                                      addOrReplaceSection( const char* SectionName,
	                                                                     const char* SectionComment = 0);
			// Removing one
			void                                      removeSection( const Section& aSection);
			void                                      removeSection( const char* SectionName);

			// SectionItem stuff: if aSectionName == 0 the default Section will be used.
			// If no default section is available an exception is thrown.
			// Testing for presence
			bool                                      containsSectionItem( const char* SectionItemName,
	                                                                     const char* SectionName = 0) const;
			// Getting it
			SectionItem&                              getSectionItem(const char* SectionItemName,
																	const char* SectionName = 0) const;
			std::string&                              getSectionItemAsString(	const char* SectionItemName,
																				const char* SectionName = 0) const;
			long                                      getSectionItemAsLong(	const char* SectionItemName,
																			const char* SectionName = 0) const;
			// Adding one. If allready present exception is thrown.
			void                                      addSectionItem(	const SectionItem& SectionItem,
																		const char* SectionName = 0);
			void                                      addSectionItem(	const char* SectionItemName,
																		const char* SectionItemValue,
																		const char* SectionItemComment = 0,
																		const char* SectionName = 0);
			void                                      addSectionItem(	const char* SectionItemName,
																		const long SectionItemValue,
																		const char* SectionItemComment = 0,
																		const char* SectionName = 0);
			// Adding or replacing (setting) one.
			void                                      addOrReplaceSectionItem(	const SectionItem& SectionItem,
																				const char* SectionName = 0);
			void                                      addOrReplaceSectionItem(	const char* SectionItemName,
																				const char* SectionItemValue,
																				const char* SectionItemComment = 0,
																				const char* SectionName = 0);
			void                                      addOrReplaceSectionItem(	const char* SectionItemName,
																				const long SectionItemValue,
																				const char* SectionItemComment = 0,
																				const char* SectionName = 0);
			// Removing one
			void                                      removeSectionItem(const SectionItem& SectionItem,
																		const char* SectionName = 0);
			void                                      removeSectionItem(const char* SectionItemName,
																		const char* SectionName = 0);
			// Misc functions
			// Rereading the profile from disk
			void                                      refresh();
			// Saving the profile to disk
			void                                      save();

			// Iterator stuff
			std::map< std::string, Section ,std::less<std::string> >::iterator             sectionsBegin();
			std::map< std::string, Section ,std::less<std::string> >::const_iterator       sectionsBegin() const;
			std::map< std::string, Section ,std::less<std::string> >::iterator             sectionsEnd();
			std::map< std::string, Section ,std::less<std::string> >::const_iterator       sectionsEnd() const;

		protected:
			std::string                                    fileName;
			std::string                                    defaultSection;
			std::string                                    comment;
			std::map< std::string, Section ,std::less<std::string> >     sections;

			void                                      readFile();
			void                                      writeFile();

		private:
			// Friends
			// This operator will put an endl into the stream as a profile is line oriented
			friend DANUIMPEXP std::ostream& operator<<(std::ostream& os,const TextProfile& aTextProfile);

	}; // class TextProfile

	DANUIMPEXP std::ostream& operator<<(std::ostream& os,const TextProfile& aTextProfile);

} // namespace Danu

#endif // TextProfile_hpp
