#include "Thread.hpp"
#include "Exception.hpp"
#include "Trace.hpp"
#include <map>
#include <iostream>

namespace Danu
{
	extern "C" void* runMemberFunction(void* aThreadFunction);
	extern "C" void* runCFunction(void* aCFunctionPtr, void* voidPointerFunctionArgument);

	/**
	 * Pthreads has no concept of threadId's (relatively small numbers that exist while
	 * the thread exists) as a Windows thread has. Because we want to use this number to
	 * distinguish threads while tracing we create them ourselves.
	 * 
	 * In this class we maintain the relation between handles and id. Id's 
	 * will be reused after it wraps at ULONG_MAX
	 */
	class Thread::ThreadIds : public std::map< Danu::ThreadHandle, Danu::ThreadId>
	{
		public:
		typedef std::map< Danu::ThreadHandle  , Danu::ThreadId>  Base;
			/**
			 * This operator returns the ThreadId for a ThreadHandle. If needed
			 * it will create a new ThreadId for this thread.
			 */
			ThreadId& operator[](const Danu::ThreadHandle& aThreadHandle)
			{
				LOLock loLock(&threadInitializer.threadIdsMutex);
				
//				if((unsigned long)aThreadHandle  == 0)
//				{
//					fprintf(stderr,"unsigned long)aThreadId ==0 ????\n");
//				}
				// If there is no ThreadId for aThreadHandle
				// a ThreadId will be created with a default ctor resulting in a
				// threadId of 0				
				ThreadId& aThreadId= ((Base*)this)->operator[](aThreadHandle);
				if(aThreadId==0) // This assumes that there will never be a valid ThreadId 0 !!!!!
				{
					// Insert the new values but first repair the faulty threadId
					((Base*)this)->operator[](aThreadHandle) = ++threadId;
				}						
				return aThreadId;
			}
			static	ThreadId threadId;
	};
	ThreadId Thread::ThreadIds::threadId;
	
	// For use in wxWindows only: the main thread may not be blocked if we want to
	// trace to the GUI. So we have to know if a thread is the main (first) thread.
	ThreadHandle mainThreadHandle = 0;

	Thread::ThreadInitializer::ThreadInitializer()
	{
		FUNCTRACE_DEVELOP();

		ThreadIds::threadId = 0;
		threadIds = new ThreadIds;				
		Trace::enableThreadId();
	}

	Thread::ThreadInitializer::~ThreadInitializer()
	{
		FUNCTRACE_DEVELOP();
		
		Trace::disableThreadId();		
		delete threadIds;		
	}
	
	//static ThreadInitializer threadInitializer;
	
	/* static */ Thread::ThreadInitializer Thread::threadInitializer;
	
	// Constructors
	Thread::Thread()
	   :  threadId(0),
	      exitCode(0)
	{
	}
	
	Thread::Thread( CFunctionPtr aCFunctionPtr,
	                void* voidPointerFunctionArgument)
	{
	   start(aCFunctionPtr,
	         voidPointerFunctionArgument);
	}
	
	Thread::Thread( AbstractThreadFunction* anAbstractThreadFunction)
	{
	   start(anAbstractThreadFunction);
	}
	
	ThreadId Thread::getThreadId() const
	{
	   if(threadId == 0)
	   {
	      // We will get here if this is actually an CurrentThread and the current thread is
		  // not created as a Thread class object
	      threadId = Thread::getCurrentThreadId();
	   }
	   return threadId;
	}
	
	// Other operations
	/* static */ ThreadId Thread::getCurrentThreadId()
	{
	   return (*threadInitializer.threadIds)[Thread::getCurrentThreadHandle()];
	}

	const ThreadHandle& Thread::getThreadHandle() const
	{
	   if(threadHandle == 0)
	   {
	      // We will get here if this is actually an CurrentThread and the current thread is
		  // not created as a Thread class object
	      threadHandle = Thread::getCurrentThreadHandle();
	   }
	   return threadHandle ;
	}
	
	ThreadHandle& Thread::getThreadHandle()
	{
	   if(threadHandle == 0)
	   {
	      // We will get here if this is actually an CurrentThread and the current thread is
		  // not created as a Thread class object
	      threadHandle = Thread::getCurrentThreadHandle();
	   }
	   return threadHandle ;
	}
	
	
	// Other operations
	/* static */ ThreadHandle Thread::getCurrentThreadHandle()
	{
	   return pthread_self();
	}
	
	/* static */ CurrentThread& Thread::current()
	{
	   static CurrentThread thread;
	   return thread;
	}
	
	void Thread::start(	CFunctionPtr aCFunctionPtr, 
						void* voidPointerFunctionArgument)
	{
		int result = pthread_create(	&threadHandle,
	  									NULL,
										aCFunctionPtr,
										voidPointerFunctionArgument);
		if(result!=0)
		{
		}
	}
	
	void Thread::start(AbstractThreadFunction* anAbstractThreadFunction)
	{
		int result = pthread_create(	&threadHandle,
										NULL,
										runMemberFunction,
										(void*)anAbstractThreadFunction);
		if(result!=0)
		{
		}
	}
	
	void Thread::stop()
	{
	}
	
	void Thread::suspend()
	{
	}
	
	void Thread::resume()
	{
	}
	
	bool Thread::isMainThread()
	{
	   return mainThreadHandle == getCurrentThreadHandle();
	}
	
	void CurrentThread::suspend()
	{
	}
	
	void CurrentThread::sleep(unsigned long aNumberOfMilliSeconds)
	{
	}
	
	void CurrentThread::waitFor(const Thread& aThread)
	{
		pthread_join(aThread.getThreadHandle(), NULL);
	}
		
	void CurrentThread::waitFor(const ThreadHandle& aThreadHandle)
	{
		pthread_join(aThreadHandle, NULL);	
	}	
	
	extern "C" void* runMemberFunction(void* aThreadFunction)
	{
		try
		{
			AbstractThreadFunction* fn = (AbstractThreadFunction*)aThreadFunction;
			fn->run();
			return 0;
		}
		catch(Danu::Exception& exception)
		{
			Danu::Trace traceObject(exception.what());
		}
		catch (std::exception& exception)
		{
			Danu::Trace traceObject(exception.what());
		}
		catch(...)
		{
			Danu::Trace traceObject("Thread: runMemberFunction unknown exception caught");
		}
		return 0;
	}

	extern "C" void* runCFunction(void* aCFunctionPtr, void* voidPointerFunctionArgument)
	{
		try
		{
			Thread::CFunctionPtr fn = (Thread::CFunctionPtr) aCFunctionPtr;
			(*fn)(voidPointerFunctionArgument);
			return 0;
		}
		catch(Danu::Exception& exception)
		{
			exception.logException();
		}
		catch (std::exception& exception)
		{
			Danu::Trace traceObject(exception.what());
		}
		catch(...)
		{
			Danu::Trace traceObject("Thread: runCFunction unknown exception caught");
		}
		return 0;
	}
	
}
