#include "Trace.hpp"
#include "TraceThreadIndention.hpp"
#include "MutexSemaphore.hpp"
#include <sys/timeb.h>    // _ftime(), time_t and struct timeb
#include <ctime>          // localtime()
#include <cstdio>         // sprintf()
#include <iostream>

namespace Danu
{	
	/* static */ Exception::ExceptionInitializer Exception::exceptionInitializer;
	/**
	 */
	Trace::TraceInitializer::TraceInitializer()
	{
		FUNCTRACE_DEVELOP();
		std::cerr << "std::cerr: + "<< __DANUFUNCTION__ << std::endl;

		// If the application is multithreaded the indentionlevel is maintained per thread
		threadIndentionLevels = new Danu::ThreadIndentionLevels;
		
		Trace::setTraceFunction(new Trace::TraceFunction);
				
		if(Trace::isExceptionTracingEnabled())
		{
			Exception::setTraceFunction(Trace::getTraceFunction());
		}
		Trace::enableTrace();
		
		std::cerr << "std::cerr: - "<< __DANUFUNCTION__ << std::endl;
	}
	/**
	 * 
	 */
	Trace::TraceInitializer::~TraceInitializer()
	{
		std::cerr << "std::cerr: + "<< __DANUFUNCTION__ << std::endl;		

		if(Trace::isExceptionTracingEnabled())
		{
			Exception::setTraceFunction(0);
		}
		Trace::disableTrace();
		
		Trace::setTraceFunction(0);
		
		delete threadIndentionLevels;
		
		std::cerr << "std::cerr: - "<< __DANUFUNCTION__ << std::endl;
	}
	
	/* static */ Trace::TraceFunction* Trace::traceFunction = 0;
	/* static */ Trace::TraceInitializer Trace::traceInitializer;
	/* static */ bool Danu::Trace::traceOn 			= true;
	/* static */ bool Danu::Trace::traceLineNumber 	= true;
	/* static */ bool Danu::Trace::traceThreadId 	= false;
	/* static */ bool Danu::Trace::traceExceptions 	= true;
	/* static */ bool Danu::Trace::traceInitialized = false;

	Trace::Trace(	const char* aText    /* = 0 */,
              		long aLineNumber     /* = 0 */,
              		bool withTimeStamp   /* = false */)
		:	traceText(aText),
			traceTimeStamp(withTimeStamp)
	{
		traceConstructor(aLineNumber);
	}

	Trace::Trace(	const std::string& aText,
              		long aLineNumber     /* = 0 */,
              		bool withTimeStamp   /* = false */)
		:	traceText(aText),
			traceTimeStamp(withTimeStamp)
	{
		traceConstructor(aLineNumber);
	}

	Trace::~Trace()
	{
		if(Trace::traceOn)
		{
			std::string timeStampString;

			// The timestamp is printed before the trace text with ":" appended
			if(traceTimeStamp==true)
			{
				timeStampString = Trace::timeStamp();
				timeStampString += ": ";
			}

			// Print the actual trace text
			std::string formattedString = timeStampString + traceText + "\n";
			Trace::traceFormattedString( formattedString, "-");
		}
	}

	/* static */ void Trace::setTraceFunction(Trace::TraceFunction* aTraceFunction /* = 0 */)
	{
		
		Trace::TraceFunction* tempFunction = Trace::traceFunction;
		Trace::traceFunction = aTraceFunction;
		delete tempFunction;
	}

	/* static */ Trace::TraceFunction* Trace::getTraceFunction()
	{
		return Trace::traceFunction;
	}


	/* static */ void Trace::enableTrace()
	{
		std::cerr << "std::cerr: + " << __DANUFUNCTION__ << std::endl;
		Trace::traceOn = true;
		TRACE_DEVELOP("Trace logging enabled\n");		
		std::cerr << "std::cerr: - " << __DANUFUNCTION__ << std::endl;
	}

	/* static */ void	Trace::disableTrace()
	{
		std::cerr << "std::cerr: + " << __DANUFUNCTION__ << std::endl;
		TRACE_DEVELOP("Trace logging disabled\n");
		Trace::traceOn = false;
		std::cerr << "std::cerr: - " << __DANUFUNCTION__ << std::endl;
	}
	/*static */ bool Trace::isTraceEnabled() 
	{
		return Trace::traceOn;
	}

	/* static */ void Trace::enableLineNumber()
	{
		Trace::traceLineNumber = true;
	}

	/* static */ void Trace::disableLineNumber()
	{
		Trace::traceLineNumber = false;
	}

	/*static*/ bool Trace::isLineNumberEnabled() 
	{
		return Trace::traceLineNumber;
	}

	/* static */ void Trace::enableThreadId()
	{
		Trace::traceThreadId = true;
	}

	/* static */ void Trace::disableThreadId()
	{
		Trace::traceThreadId = false;
	}
	/* static */ bool Trace::isThreadIdEnabled()
	{
		return Trace::traceThreadId;
	}

	/* static */ void Trace::enableExceptionTracing()
	{
		Trace::traceExceptions = true;
		// Now we have a trace function we replace the default (simple)
		// exception trace function with the more versatile trace trace function.
		Exception::setTraceFunction(Trace::getTraceFunction());
	}

	/* static */ void Trace::disableExceptionTracing()
	{
		Trace::traceExceptions = false;
		// This will reinstate the default (simple) exception function.
		Exception::setTraceFunction(0);
	}

	/*static */ bool Trace::isExceptionTracingEnabled() 
	{
		return Trace::traceExceptions;
	}

	/* static */ void Trace::trace(	const std::string& aText,
									bool withTimeStamp /* = Trace::isTraceEnabled() */)
	{
		std::string textString(aText);
		Trace::traceFormattedString(textString,">");
	}

	/* static */ void Trace::trace(	const char* aText,
									bool withTimeStamp /* = Trace::isTraceEnabled() */)
	{
		std::string textString(aText);
		Trace::trace(textString,withTimeStamp);
	}

	//	The two functions above resolve to this function after formatting the text
	/* static */ void Trace::traceFormattedString(	const std::string& aText,
													char* aTraceMarker)	// 	This text will be prepended
																		// 	After the text a newline will be inserted
																		//	as approriate
	{
		if(Trace::traceOn && Trace::traceFunction)
		{
			LOLock lock(&traceInitializer.traceFunctionMutex);

   			std::string margin;
   			std::string traceText;

			if(Trace::traceThreadId)
			{
				ThreadHandle currentThreadHandle = Thread::getCurrentThreadHandle();
				ThreadId currentThreadId = Thread::getCurrentThreadId();

				switch(*aTraceMarker)
				{
					case '+':
					{
						ThreadIndentionLevel level = (*traceInitializer.threadIndentionLevels)[currentThreadHandle];
						++level;						
						margin.insert((unsigned int)0,level.indentionLevel,' ');
						(*traceInitializer.threadIndentionLevels)[currentThreadHandle] = level;
						break;
					}
					case '-':
					{
						ThreadIndentionLevel level = (*traceInitializer.threadIndentionLevels)[currentThreadHandle];						
						if( level.indentionLevel < 0) // Unbalanced +/-
						{
							margin.insert((unsigned int)0,0,' ');
						}else
						{
							margin.insert((unsigned int)0,level.indentionLevel ,' ');
							--level;
							(*traceInitializer.threadIndentionLevels)[currentThreadHandle] = level;
						}
						break;
					}
					default:
					{
						margin.insert((unsigned int)0,(*traceInitializer.threadIndentionLevels)[currentThreadHandle].indentionLevel + 3,' ');
						break;
					}
				}
				traceText = strprintf("%s%lu: %c %s",margin.c_str(),currentThreadId,*aTraceMarker,aText.c_str());		
			}else
			{
				traceText = strprintf("%s%c %s",margin.c_str(),*aTraceMarker,aText.c_str());
			}
			traceFunction->trace(traceText .c_str());
		}
	}

	/* static */ std::string Trace::timeStamp()
	{
		// \todo ftime and localtime are not threadsafe????
   		char timeStamp[27];
   		struct timeb timeBuffer;
   		ftime(&timeBuffer);
   		struct tm* timeStruct = localtime(&timeBuffer.time);
   		sprintf(timeStamp,"%04d-%02d-%02d-%02d-%02d-%02d-%06d",timeStruct->tm_year+1900,timeStruct->tm_mon+1,timeStruct->tm_mday,timeStruct->tm_hour,timeStruct->tm_min,timeStruct->tm_sec,timeBuffer.millitm);
   		return &timeStamp[0];
	}
	void Trace::traceConstructor(long aLineNumber)
	{
		if(Trace::traceOn)
		{
			if (traceText.length() > 0)
			{
				std::string timeStampString;
				std::string lineNumberString;

				if(traceTimeStamp==true)
				{
					timeStampString = Trace::timeStamp();
					timeStampString += ": ";
				}

				if(Trace::traceLineNumber)
				{
					lineNumberString = " (line ";
					// max long ==__LONG_MAX__ == 2147483647 == 10 chars
					char numberBuffer[10+1];
					sprintf(&numberBuffer[0],"%ld",aLineNumber);
					lineNumberString += std::string(&numberBuffer[0]);
					lineNumberString += ")";
				}
				std::string formattedString = timeStampString + traceText + lineNumberString + "\n";
				Trace::traceFormattedString( formattedString, "+");
			} else
			{
				Trace::traceFormattedString("??? No Text ???", "+");
			}
		}
	}
} // namespace Danu

